from os import path
from shutil import rmtree

import pytest
from distributed import Client, LocalCluster
from mason_dask.utils.cluster_spec import ClusterSpec

from mason_dask.definitions import from_root
from mason_dask.jobs.executed import InvalidJob
from mason_dask.jobs.query import QueryJob, ValidQueryJob
from dask import dataframe as dd

TMP = "../tmp/"
TEST_DATA = from_root("/test/data/test_data.csv")

@pytest.fixture(autouse=True)
def run_around_tests():
    yield
    clear_tmp()

@pytest.fixture(scope="module")
def client():
    with Client(LocalCluster(n_workers=2)) as client:
        yield(client)

def clear_tmp():
    if path.exists(TMP):
        rmtree(TMP)

def test_schema(client):
    in_paths = ["good_input_path", "good_input_path2"]
    in_format = "csv"
    out_path = "good_output_path"
    query_string = "SELECT * FROM table LIMIT 3"

    good_spec = {
        "input_paths": in_paths,
        "input_format": in_format,
        "query_string": query_string,
        "output_path": out_path
    }

    job = QueryJob(good_spec).validate()
    assert (isinstance(job, ValidQueryJob))
    
    bad_spec = {
        "input_paths": "1234",
        "input_format": "avro",
        "query_string": "1234",
        "output_path": "1234",
    }

    for key, value in bad_spec.items():
        good_spec[key] = value
        job = QueryJob(good_spec).validate()
        assert (isinstance(job, InvalidJob))

def test_basic(client):

    cluster_spec = ClusterSpec(client)
    spec = {
        "input_paths": [TEST_DATA],
        "input_format": "csv",
        "output_path": TMP + "query_out/",
        "query_string": "SELECT * from $table LIMIT 3"
    }

    job = QueryJob(spec).validate()
    assert (isinstance(job, ValidQueryJob))
    assert(job.run(cluster_spec).bind(lambda e: e.message) == "Table succesfully formatted as parquet and exported to ../tmp/query_out/") # type: ignore
    df = dd.read_parquet(TMP + f"/query_out/part.0.parquet").compute()

    assert(df.shape[0] == 3)
    assert(sorted(list(df["col_a"])) == [123.0, 123.0, 789.0])
    assert(sorted(list(df["col_b"])) == ["test2", "test2", "test3"])
    assert(sorted(list(df["col_c"])) == [456.0, 456.0, 456.0])

    # where
    spec = {
        "input_paths": [TEST_DATA],
        "input_format": "csv",
        "output_path": TMP + "query_out/",
        "query_string": "SELECT * from $table WHERE col_a <= 123.0"
    }
    job = QueryJob(spec).validate()
    assert (isinstance(job, ValidQueryJob))
    job.run(cluster_spec)

    assert (isinstance(job, ValidQueryJob))
    df = dd.read_parquet(TMP + f"/query_out/part.0.parquet").compute()

    assert (df.shape[0] == 2)
    assert (sorted(list(df["col_a"])) == [123.0, 123.0])
    assert (sorted(list(df["col_b"])) == ["test2", "test2"])
    assert (sorted(list(df["col_c"])) == [456.0, 456.0])







from collections import namedtuple
from functools import partial
from typing import List, Union, Optional

from dask.dataframe import DataFrame
from dask import dataframe as dd
from returns.result import Result, Success, Failure
from returns.converters import flatten
from returns.pipeline import flow
from returns.pointfree import bind
from schema import Schema, Use, And, SchemaError
from schema import Optional as SOptional
from pandas import DataFrame as PDataFrame

from .valid_job import ValidJob
from .executed import InvalidJob, ExecutedJob
from ..models.dataframe_from import VALID_INPUT_FORMATS, df_from
from ..models.dataframe_to import df_to
from ..utils.cluster_spec import ClusterSpec
from ..utils.exception import message

VALID_OUTPUT_FORMATS = ["csv", "json", "xlsx", "parquet"]

class FormatJob:
    
    def __init__(self, spec: dict):
        self.spec = spec

    def schema(self) -> Schema:
        input_formats = And(Use(str), lambda n: n in VALID_INPUT_FORMATS)
        output_formats = And(Use(str), lambda n: n in VALID_OUTPUT_FORMATS)

        schema = {
            "input_paths": [Use(str)],
            "input_format": input_formats,
            "output_format": output_formats,
            "output_path": Use(str),
            "partition_columns": [Use(str)],
            "filter_columns": [Use(str)],
            SOptional("credentials"): Use(dict),
            SOptional("line_terminator", default="\n"): str,
            SOptional("partitions"): Use(int)
        }
        return Schema(schema)

    def validate(self) -> Union['ValidFormatJob', InvalidJob]:
        try:
            d = self.schema().validate(self.spec)
            t = namedtuple("JobAttributes", d.keys())(*d.values())
            return ValidFormatJob(t)
        except SchemaError as e:
            return InvalidJob(f"Invalid Schema: {message(e)}")

class ValidFormatJob(ValidJob):

    def __init__(self, t):
        self.input_paths: List[str] = t.input_paths
        self.input_format: str = t.input_format
        self.output_format: str = t.output_format
        outpath: str = t.output_path
        
        if outpath.endswith("/"):
            self.output_path = outpath
        else:
            self.output_path = outpath + "/"
            
        self.partition_columns: List[str] = t.partition_columns
        self.filter_columns: List[str] = t.filter_columns
        self.line_terminator: str = t.line_terminator
        self.partitions: Optional[int]
        try:
            self.partitions = t.partitions
        except AttributeError:
            self.partitions = None 

    def check_columns(self, df: DataFrame, columns: List[str]) -> Result[DataFrame, InvalidJob]:
        keys = df.dtypes.keys()
        diff = set(columns).difference(set(df.dtypes.keys()))
        if len(diff) == 0:
            return Success(df)
        else:
            return Failure(InvalidJob(f"Filter columns {', '.join(diff)} not a subset of {', '.join(keys)}"))

    def repartition(self, df: DataFrame, cluster_spec: ClusterSpec) -> Result[DataFrame, InvalidJob]:
        size = df.size.compute()
        
        num_partitions = self.partitions

        if cluster_spec.valid():
            cluster_spec_num_workers = cluster_spec.num_workers() 
        else:
            cluster_spec_num_workers = None
            
        num_workers = num_partitions or cluster_spec_num_workers 
        
        if num_workers:
            max_partition_size = cluster_spec.max_partition_size()
            even_partition_size = size / num_workers 
            if max_partition_size:
                if even_partition_size > max_partition_size:
                    return Failure(InvalidJob(f"Partitions too large for workers: (size, max) = ({even_partition_size}, {max_partition_size}).  Add more workers or increase worker memory."))
                else:
                    return Success(df.repartition(partition_size=even_partition_size).repartition(npartitions=num_workers))
            else:
                return Success(df.repartition(partition_size=even_partition_size).repartition(npartitions=num_workers))
        else:
            return Success(df)


    def df_filter_columns(self, cluster_spec: ClusterSpec, df: DataFrame) -> Result[DataFrame, InvalidJob]:
        if len(self.filter_columns) > 0:
            check = self.check_columns(df, self.filter_columns)
            
            def filter_and_repartition(df: DataFrame) -> Result[DataFrame, InvalidJob]:
                filtered = df[self.filter_columns]
                repartitioned = self.repartition(filtered, cluster_spec)
                return repartitioned
                
            return flatten(check.map(lambda b: filter_and_repartition(b)))
        else:
            return Success(df)


    def partition_and_write(self, cluster_spec: ClusterSpec, df: DataFrame) -> Result[ExecutedJob, InvalidJob]:
        if len(self.partition_columns) > 0:
            def write_partitioned(df: PDataFrame) -> Result[ExecutedJob, InvalidJob]:
                ddf = dd.from_pandas(df, npartitions=cluster_spec.num_workers() or self.partitions or 10)
                labels = {}

                for p in self.partition_columns:
                    value = df[p].unique()[0]
                    labels[p] = str(value)

                label = "&".join(list(map(lambda item: "=".join(item), labels.items())))
                result = df_to(ddf, self.output_path, self.output_format, label)

                return result

            def repartition_and_write(df: DataFrame) -> Result[ExecutedJob, InvalidJob]:
                pc = self.partition_columns
                df.groupby(pc).apply(write_partitioned, meta=str).compute()
                return Success(ExecutedJob(f"Repartitioned dataframe according to columns {self.partition_columns} and output to {self.output_path}"))
            
            check = self.check_columns(df, self.partition_columns)
            return flatten(check.map(lambda b: repartition_and_write(b)))
        else:
            repartitioned = self.repartition(df, cluster_spec)
            return flatten(repartitioned.map(lambda b: df_to(b, self.output_path, self.output_format)))

    def run(self, cluster_spec: ClusterSpec) -> Result[ExecutedJob, InvalidJob]:
        do_filter_columns = partial(self.df_filter_columns, cluster_spec)
        do_partition_and_write = partial(self.partition_and_write, cluster_spec)

        final = flow(
            df_from(self.input_paths, self.input_format, self.line_terminator),
            bind(do_filter_columns),
            bind(do_partition_and_write)
        )
        
        return final




from mason_dask.utils.exception import message
from returns.result import Result, Success, Failure
from returns.pointfree import bind
from returns.pipeline import flow
from functools import partial

def test_nested_pipe():
    class TestClass():

        def __init__(self, i: int):
            self.i = i

        def regular_function(self, arg: int) -> float:
            return float(arg)

        def returns_container(self, test: float, arg: float) -> Result[str, ValueError]:
            if arg == test:
                return Success(str(arg))
            else:
                return Failure(ValueError('Wrong arg'))

        def also_returns_container(self, arg: str) -> Result[str, ValueError]:
            if arg != "0":
                return Success(str(arg))
            else:
                return Failure(ValueError('Wrong arg 2'))

        def also_returns_container_2(self, arg: str) -> Result[str, ValueError]:
            return Success(arg + '!')

        def run(self):
            a = flow(
                self.i,
                self.regular_function,
                partial(self.returns_container, 0.0),
                bind(self.also_returns_container),
                bind(self.also_returns_container_2)
            )
            return a

    assert(TestClass(0).run()._inner_value == "0.0!")
    assert(message(TestClass(1).run()._inner_value) == "ValueError: Wrong arg\n")





from typing import List
from dask.dataframe import DataFrame
from mason_dask.jobs.executed import InvalidJob
from dask import dataframe as dd
from returns.result import Result, Failure

from mason_dask.utils.returns import safe_obj

VALID_TEXT_FORMATS = ["csv", "csv-crlf"]
VALID_JSON_FORMATS = ["json", "jsonl"]
VALID_INPUT_FORMATS = VALID_TEXT_FORMATS + VALID_JSON_FORMATS + ["parquet"]

def df_from(input_paths: List[str], input_format: str, line_terminator: str) -> Result[DataFrame, InvalidJob]:
    df: DataFrame
    if input_format in VALID_TEXT_FORMATS:
        @safe_obj
        def read_csv(input_paths, line_terminator, assume_missing, sample) -> DataFrame:
            return dd.read_csv(input_paths, lineterminator=line_terminator, assume_missing=assume_missing, sample=sample)
        return read_csv(input_paths, line_terminator, True, 25000000)
    elif input_format == "parquet":
        @safe_obj
        def read_parquet(input_paths) -> DataFrame:
            return dd.read_parquet(input_paths)
        return read_parquet(input_paths)
    elif input_format in VALID_JSON_FORMATS:
        @safe_obj
        def read_json(input_paths) -> DataFrame:
            return dd.read_json(input_paths)
        return read_json(input_paths) 
    else:
        return Failure(InvalidJob("Invalid Input Format"))



from abc import abstractmethod

from mason_dask.jobs.executed import ExecutedJob, InvalidJob
from returns.result import Result
from mason_dask.utils.cluster_spec import ClusterSpec


class ValidJob:
    
    @abstractmethod
    def run(self, cluster_spec: ClusterSpec) -> Result[ExecutedJob, InvalidJob]:
        raise NotImplementedError("run method not implemented")



from functools import wraps

from mason_dask.utils.exception import message

from mason_dask.jobs.executed import ExecutedJob, InvalidJob
from returns.result import Result, Success, Failure


def safe_job(func):
    @wraps(func)
    def wrapper(*args, **kwargs) -> Result[ExecutedJob, InvalidJob]:
        try:
            result = func(*args, **kwargs)
            return Success(ExecutedJob(result))
        except Exception as e:
            return Failure(InvalidJob(message(e)))
    return wrapper

def safe_obj(func):
    @wraps(func)
    def wrapper(*args, **kwargs) -> Result[ExecutedJob, InvalidJob]:
        try:
            return Success(func(*args, **kwargs))
        except Exception as e:
            return Failure(InvalidJob(message(e)))
    return wrapper
